/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * 
 * (c) kernel concepts 2006, 2007
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Application launcher user interface.
 */
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <libintl.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <gpe/spacing.h>
#include <gpe/infoprint.h>
#include <gpe/launch.h>
#include <gpe/desktop_file.h>
#include <gpe/pixmaps.h>
#include <gpe/gtkappnotification.h>
#include <gpe/settings.h>

#include "homescreen.h"
#include "applications.h"
#include "gpe-applauncher.h"
#include "dbus-if.h"
#include "launch-app.h"

#define _(x) gettext (x)

/* Number of displayed columns in icon and list layout. */
#define COLUMNS_STACK   3
#define COLUMNS_SIDE    1


#define KEY_BG_IMAGE "/apps/mainmenu/backgroundimage"
#define KEY_FOLDERS_PATH "/apps/mainmenu/folderspath"
#define KEY_APPS_PATH "/apps/mainmenu/appspath"
#define KEY_ITEM_LAYOUT "/apps/mainmenu/layout"
#define KEY_THEME_BG "/mainmenu/backgroundimage/name"
#define KEY_THEME_ICON_SIZE_NORMAL  "/mainmenuitem/icon/normal/size"
#define KEY_THEME_ICON_SIZE_FOCUSED "/mainmenuitem/icon/focused/size"

#define DEFAULT_FOLDERS_PATH PREFIX "/share/gpe/vfolders" 
#define DEFAULT_THEME_BG PREFIX  "/share/themes/default/homescreen_background.png" 
#define DEFAULT_THEME_ICON_SIZE_NORMAL  32
#define DEFAULT_THEME_ICON_SIZE_FOCUSED 36

/* Data type for application folder/tab encapsulation. */
typedef struct {
    gint pagenum;         /* pagenum of the notebook tab */
    gchar *name;          /* name / label */
    gchar *iconname;      /* icon filename */
    gchar *category;      /* matching category, maybe wants to be extended to multiple */
    GtkWidget *icon;      /* icon widget */
    GtkTreeModel *storage;/* associated application item data */
    GtkWidget *iconview;  /* widget used to display data */
} t_folder;

/* Application property fields for list store. */
typedef enum {
    APROP_NAME,
    APROP_COMMENT,
    APROP_EXEC,
    APROP_TERMINAL,
    APROP_STARTUP,
    APROP_SINGLE,
    APROP_ICONNAME,
    APROP_ICONDATA,
    APROP_TYPE,
    APROP_MAX
} t_app_property;


static GList *folderlist;
static t_folder *default_folder = NULL;
#ifdef GRIDLAYOUT
static t_folder *super_folder = NULL;
#endif


#ifdef GRIDLAYOUT
static gint 
get_pagenum (const gchar * name)
{
  GList *iter;
  gint pagenum = 0;
	
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      if (strcmp(name, folder->name) == 0)
        {
          pagenum = folder->pagenum;
          break;
        }
    }
  return pagenum;
}

static GtkWidget *
retrieve_notebook (void)
{
    GList *iter = folderlist;
    if (!iter)
      return NULL;
    t_folder *folder = iter->data;
    if (folder)
    {
         GtkWidget *parent = folder->iconview;
         while (!GTK_IS_NOTEBOOK (parent) && parent != NULL)
           {
             parent = gtk_widget_get_parent (parent);
           }
         if (GTK_IS_NOTEBOOK (parent) && parent != NULL)
           return parent;
    }
    return NULL;
}
#endif


/* Update stored application pixmaps in order to support setting changes. */
static gboolean
update_item_pixmap (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
                    gpointer data)
{
  GdkPixbuf *pbuf = NULL;
  gchar *iconname = NULL;
  gint size = (gint)data;  
    
  gtk_tree_model_get (model, iter, 
                      APROP_ICONDATA, &pbuf, APROP_ICONNAME, &iconname, -1);
  if (pbuf)
    g_object_unref (pbuf);
  
  pbuf = load_icon_scaled (iconname, size);
  
  if (!pbuf)
    {
      pbuf = gpe_find_icon ("applauncher");
      pbuf = gdk_pixbuf_scale_simple (pbuf, size, size, GDK_INTERP_BILINEAR);
    }
  
  gtk_list_store_set (GTK_LIST_STORE (model), iter, 
                      APROP_ICONDATA, pbuf, -1);
    
  if (pbuf)
    g_object_unref (pbuf);
 
  return FALSE;
}

/* Handle theme changes. */
void 
applications_update_theme (const gchar *themename)
{
  GList *iter;
  gchar *resource_path = NULL;
  gchar *bg_file = NULL;	
  gchar *key;
    
  if (themename == NULL)
    return;
    
  /*
   * set the icon of each application
   */   
  setting_get_string (setting, KEY_RESOURCE_PATH, &resource_path);
  if (!resource_path)
	  resource_path = g_strdup (DEFAULT_RESOURCE_PATH);
  
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      GdkPixbuf *pbuf = NULL, *iconbuf = NULL;
      gchar *str = NULL;

      str = g_strconcat (resource_path, themename, "/", folder->iconname, NULL);      
      pbuf = gdk_pixbuf_new_from_file (str, NULL); 

      g_free (str);
      if (pbuf == NULL)
        continue;
      
      iconbuf = gdk_pixbuf_scale_simple (pbuf, TAB_ICON_SIZE, TAB_ICON_SIZE, GDK_INTERP_BILINEAR);
      gtk_image_set_from_pixbuf (GTK_IMAGE (folder->icon), iconbuf);
      g_object_unref (pbuf);
      g_object_unref (iconbuf);
    }
  g_free (resource_path);
  
  /* 
   * set background image for every folder 
   */
  key = g_strconcat (THEME_PREFIX, themename, KEY_THEME_BG, NULL);  
  setting_get_string (setting, key, &bg_file);
  g_free (key);
  if (!bg_file)
      bg_file = g_strdup (DEFAULT_THEME_BG);
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      render_bg (bg_file, folder->iconview);
    }
  g_free (bg_file);
}

/* Change background image of application area - currently unsupported. */
void 
applications_update_background (GtkWidget *window)
{
  GList *iter;
    
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
                  
      gdk_window_set_back_pixmap (folder->iconview->window,  
                                  window->style->bg_pixmap[GTK_STATE_NORMAL],
                                  FALSE);
      gtk_widget_queue_draw (folder->iconview);
    }
}

/* Change the size of the application icons. */
void
applications_update_itemsize (gint width)
{
  GList *iter;

  gint colwidth;
  gint layout = -1;
  
  setting_get_int (setting, KEY_ITEM_LAYOUT, &layout);
  if (layout < LAYOUT_STACK)
      layout = LAYOUT_STACK;
    
  colwidth = (gdk_screen_width() - 20) / (layout ? COLUMNS_SIDE : COLUMNS_STACK);
	
  if (layout == LAYOUT_SIDE)
      width /= 2;
    
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      gtk_icon_view_set_item_width (GTK_ICON_VIEW(folder->iconview), colwidth);
      gtk_tree_model_foreach (GTK_TREE_MODEL(folder->storage), update_item_pixmap, (gpointer) width);
    }
}

static gint 
get_item_size (gint status)
{
  gint size = 0;
  gchar *current_theme = NULL;
  gchar *key;
  
  setting_get_string (setting, KEY_CURRENT_THEME, &current_theme);
  if (current_theme == NULL)
    current_theme = g_strdup (DEFAULT_THEME);

  switch (status)
  {
  	case ITEM_NORMAL:
      key = g_strconcat (THEME_PREFIX, current_theme, KEY_THEME_ICON_SIZE_NORMAL, NULL);
      break;
    case ITEM_FOCUSED:  
      key = g_strconcat (THEME_PREFIX, current_theme, KEY_THEME_ICON_SIZE_FOCUSED, NULL);
      break;
    default:
      key = g_strconcat (THEME_PREFIX, current_theme, KEY_THEME_ICON_SIZE_NORMAL, NULL);
  }
  
  setting_get_int (setting, key, &size);
  if (!size)
     size = (status == ITEM_FOCUSED) ?  DEFAULT_THEME_ICON_SIZE_FOCUSED 
                                      : DEFAULT_THEME_ICON_SIZE_NORMAL;
  g_free (key);  
  g_free (current_theme);
  
  return size;	
}

/* Change the layout of the application displays. */
void
applications_update_item_layout (gint type)
{
  GList *iter;
  GtkOrientation or;
  gint size;
    
  switch (type)
    {
        case LAYOUT_STACK:
            or = GTK_ORIENTATION_VERTICAL;
        break;
        case LAYOUT_SIDE:
            or = GTK_ORIENTATION_HORIZONTAL;
        break;
        default:
            or = GTK_ORIENTATION_VERTICAL;
    }

  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      gtk_icon_view_set_orientation (GTK_ICON_VIEW(folder->iconview), or);
      gtk_icon_view_set_columns (GTK_ICON_VIEW (folder->iconview), 
                                 type ? COLUMNS_SIDE : COLUMNS_STACK);
    }
    
  size = get_item_size (ITEM_NORMAL);    
  applications_update_itemsize (size);
}


int 
applications_start_service (const gchar *service_name, const gchar *method, 
							const gchar* caller_service_name)
{
  GList *iter;
  gchar *name = NULL;
  gchar *exec = NULL;
  gboolean single_instance = TRUE;
  gboolean startup_notify = TRUE;
  gboolean ret;	
  gint out = 0;

#ifdef DEBUG
  if (caller_service_name != NULL)
    g_message ("caller is %s", caller_service_name);
#endif
  
  for (iter = folderlist; iter; iter = iter->next)
    {
      if (out == 1)
          break;
      
      t_folder *folder = iter->data;
      GtkTreeIter *tree_iter;                
      GtkTreeIter first_iter;          
              
      if (gtk_tree_model_get_iter_first (folder->storage, &first_iter))
        tree_iter = &first_iter;
      else /* an empty folder */
        continue;
      do
        {
          gtk_tree_model_get (folder->storage, tree_iter, APROP_NAME, &name, -1);			       
          if (strcmp (service_name, name) == 0)
            {
              gtk_tree_model_get (folder->storage, tree_iter, APROP_EXEC, &exec, 
                                  APROP_SINGLE, &single_instance,					   						 
                                  APROP_STARTUP, &startup_notify, -1);
              out = 1;
              break;
            }   
          if (name) {g_free (name); name = NULL;}
        } 
      while (gtk_tree_model_iter_next (folder->storage, tree_iter));
	}
	
#ifdef DEBUG
  g_message ("%s is %s\n", 
						 (service_name ? service_name : "(null)"),
						 (exec ? exec : "(null)"));
#endif
    
  ret = launch_application (name, exec, method, single_instance, 
                            startup_notify, caller_service_name);

  if (exec) 
      g_free (exec); 
  if (name) 
      g_free (name); 
	
  return (ret ? 0 : 1);
}
	
/* Start an application honoring single instance applications and notification. */
static void
applications_activate (GtkIconView *iconview, GtkTreePath *path, gpointer user_data)
{
  t_folder *folder = (t_folder*)user_data;
  GtkTreeIter iter;
  gchar *exec;
  gchar *name;
  gchar *type;
  gboolean single_instance = TRUE;
  gboolean startup_notify;
  gchar *text;
  Display *dpy = GDK_DISPLAY();
    
  gtk_tree_model_get_iter (folder->storage, &iter, path);
  gtk_tree_model_get(folder->storage, &iter, APROP_EXEC, &exec, 
                                             APROP_SINGLE, &single_instance,
                                             APROP_STARTUP, &startup_notify,
                                             APROP_NAME, &name,
                                             APROP_TYPE, &type, -1);
  if (strcmp (type, "Directory") == 0)
  {
       GList *iter;
       gint pagenum = 0;
    for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      if (strcmp(name, folder->name) == 0)
      {
        pagenum = folder->pagenum;
        break;
      }
    }
    GtkWidget *parent = folder->iconview;
    while (!GTK_IS_NOTEBOOK (parent) && parent != NULL)
      {
        parent = gtk_widget_get_parent (parent);
      }
    if (GTK_IS_NOTEBOOK (parent) && parent != NULL)
      gtk_notebook_set_current_page (GTK_NOTEBOOK (parent), pagenum);
  }
  else
  {  
    if (name == NULL) 
      name = g_strdup (exec);

    text = g_strdup_printf ("%s %s", _("Starting"), name);
    gpe_popup_infoprint (dpy, text);
    g_free (text);
  
    launch_application (name, exec, "", single_instance, startup_notify, 
                      SERVICE_APPLAUNCHER);
  }	
  if (exec) g_free (exec); 
  if (name) g_free (name); 
  if (type) g_free (type);
  
  return;
}

/* Create a new application folder widget from its desktop description. */
static t_folder*
new_folder_from_desktop (GnomeDesktopFile *g)
{
  t_folder *result = g_malloc0 (sizeof(t_folder));  
  GdkPixbuf *iconbuf;
  gchar *str;
	
  gnome_desktop_file_get_string (g, NULL, "Name", &str);
#ifndef GRIDLAYOUT
  if (!str || !strcmp (str, _("Super")))
    {
      g_free (result);
      g_free (str);
      return NULL;
    }
#endif
  result->name = str;
  gnome_desktop_file_get_string (g, NULL, "Icon", &str);
  result->iconname = str;
  gnome_desktop_file_get_string (g, NULL, "Match", &str);
  result->category = str; 
  
  iconbuf = load_icon_scaled (result->iconname, TAB_ICON_SIZE);

  result->icon = gtk_image_new_from_pixbuf (iconbuf);
  gtk_widget_set_usize (result->icon, (gdk_screen_width() - 42) / 4, -1);
    
  g_object_unref (iconbuf);
  
  /* is this our default folder? */
  if (!strcmp("Other", result->category))
      default_folder = result;
#ifdef GRIDLAYOUT
  else if (!strcmp("SUPER", result->category))
      super_folder = result;
#endif
  return result;
}

/* Create a new allpication data set from its desktop description and add this
 * to the list store. */
static void
new_application_from_desktop (GnomeDesktopFile *g, GtkTreeModel *store)
{
  GtkTreeIter iter;
  gchar *name, *icon, *exec, *type;
  gboolean terminal;
  gboolean startup;
  gboolean single = TRUE;
  gboolean bv;
  GdkPixbuf *ibuf;
  gint size;
    
  gnome_desktop_file_get_string (g, NULL, "Name", &name);
  gnome_desktop_file_get_string (g, NULL, "Icon", &icon);
  gnome_desktop_file_get_string (g, NULL, "Exec", &exec);
  gnome_desktop_file_get_string (g, NULL, "Type", &type);
  if (gnome_desktop_file_get_boolean (g, NULL, "Terminal", &terminal) == FALSE)
      terminal = FALSE;
  if (gnome_desktop_file_get_boolean (g, NULL, "StartupNotify", &startup) == FALSE)
      startup = TRUE;
  if (gnome_desktop_file_get_boolean (g, NULL, "SingleInstance", &bv) && (bv == FALSE))
      single = FALSE;
  if (gnome_desktop_file_get_boolean (g, NULL, "X-SingleInstance", &bv) && (bv == FALSE))
      single = FALSE;

  size = get_item_size (ITEM_NORMAL);
  ibuf = load_icon_scaled (icon, size);
  
  if (!ibuf)
    {
      ibuf = gpe_find_icon ("applauncher");
      ibuf = gdk_pixbuf_scale_simple (ibuf, size, size, GDK_INTERP_BILINEAR);	
    }
  
  if (ibuf)
    {
      gtk_list_store_append (GTK_LIST_STORE (store), &iter);
      gtk_list_store_set (GTK_LIST_STORE (store), &iter, APROP_NAME, name,
                          APROP_ICONDATA, ibuf, APROP_ICONNAME, icon, 
                          APROP_EXEC, exec, 
                          APROP_TERMINAL, terminal, APROP_STARTUP, startup, 
                          APROP_SINGLE, single, APROP_TYPE, type, -1);
      g_object_unref (ibuf);
    }
  g_free (name);
  g_free (icon);
  g_free (type);
  g_free (exec);
}


/* Load all available desktop descriptions of folders. */
static GList*
load_folder_list (const gchar *path)
{
  DIR *dir;
  struct dirent *entry;
  GList *result = NULL;
  t_folder *new_folder;

  if (path == NULL)
    return NULL;
    
  dir = opendir (path);
  if (dir)
    {
      while ((entry = readdir (dir)))
        {
	      gchar *temp;
	      GnomeDesktopFile *p;
	      GError *err = NULL;
	  
          if (entry->d_name[0] == '.')
            continue;
        
          temp = g_strdup_printf ("%s/%s", path, entry->d_name);
          p = gnome_desktop_file_load (temp, &err);
          if (p)
            {
              gchar *type;
              gnome_desktop_file_get_string (p, NULL, "Type", &type);
    
              if (type == NULL || strcmp (type, "Directory"))
		{
                  gnome_desktop_file_free (p);
		}
              else
                {
                  new_folder = new_folder_from_desktop (p);
                  if (new_folder)
                      result = g_list_append (result, new_folder);
                  gnome_desktop_file_free (p);
                }
    
              if (type)
                 g_free (type);
            }
          else
            g_printerr ("Couldn't load \"%s\": %s\n", temp, err->message);
    
          if (err)
            g_error_free (err);
    
          g_free (temp);
      }
    closedir (dir);
  }
  return result;
}

/* Load all available application descriptions. */
void
load_application_list (const gchar *path)
{
  DIR *dir;
  struct dirent *entry;

  g_return_if_fail (path);
  
  dir = opendir (path);
  if (dir)
    {
      while ((entry = readdir (dir)))
        {
          char *temp;	  
      
          if ((entry->d_name[0] == '.') || (!g_strrstr (entry->d_name, ".desktop")))
              continue;
    
          temp = g_strdup_printf ("%s/%s", path, entry->d_name);
#ifdef DEBUG
          g_message ("load app: %s", entry->d_name);
#endif          
          add_application (temp);	 
     
          g_free (temp);
        }
      closedir (dir);	 
    }
}

/* Add a new application description to a folder */
void
add_application (const gchar *filename)
{
  GnomeDesktopFile *p;
  GError *err = NULL;
  p = gnome_desktop_file_load (filename, &err);
  if (p)
    {
      gchar *type;
      gnome_desktop_file_get_string (p, NULL, "Type", &type);

      if (type == NULL)
	{
          gnome_desktop_file_free (p);
	  return;
	}
      if (!strcmp (type, "PanelApp"))
	{
          gnome_desktop_file_free (p);
          g_free (type);
	  return;
	}
#ifdef GRIDLAYOUT			
      /* handle folders */
      if ((super_folder != NULL) && !strcmp (type, "Directory")) 
        {
	  gchar *name;
          gnome_desktop_file_get_string (p, NULL, "Name", &name);
	  if (strcmp (name, _("Super"))) //we should avoid super folder being added to itself
              new_application_from_desktop (p, super_folder->storage);
          g_free (name);
          g_free (type);
          gnome_desktop_file_free (p);
	  return;
	}
#endif
      else
        {
          GList *iter;
          gchar *cat;			
          gboolean match = FALSE;
          gnome_desktop_file_get_string (p, NULL, "Categories", &cat);
          for (iter = folderlist; iter; iter = iter->next)
            {
              t_folder *folder = iter->data;
              if (cat && strstr(cat, folder->category))
                {
                  new_application_from_desktop (p, folder->storage);
                  match = TRUE;
                }
            }
          if ((match == FALSE) && (default_folder != NULL)) /* no folder found */
              new_application_from_desktop (p, default_folder->storage);
                  
          g_free (cat);
        }
      if (type)
          g_free (type);
    }
  else
     fprintf (stderr, "Couldn't load \"%s\": %s\n", filename, err->message);

  gnome_desktop_file_free (p);
  if (err)
     g_error_free (err);
}

/* Remove a new application description from a folder*/
void
remove_application (const gchar *filename)
{
  GnomeDesktopFile *p;
  GError *err = NULL;
  p = gnome_desktop_file_load (filename, &err);
  if (p)
    {
      gchar *type;
      gnome_desktop_file_get_string (p, NULL, "Type", &type);

      if (type == NULL || strcmp (type, "Application"))
          gnome_desktop_file_free (p);
      else
        {
          GList *iter;
          gchar *name;              
          gchar *cat;              
          gnome_desktop_file_get_string (p, NULL, "Categories", &cat); 
          gnome_desktop_file_get_string (p, NULL, "Name", &name);
          for (iter = folderlist; iter; iter = iter->next)
            {
              t_folder *folder = iter->data;
              if (cat && strstr(cat, folder->category))
                {
                  GtkTreeIter tree_iter;                
                  gchar *tree_iter_name;
                  gboolean end_reached;
 
                  end_reached = gtk_tree_model_get_iter_first (folder->storage, &tree_iter);
                  while (!end_reached)
                    {
                      gtk_tree_model_get(folder->storage, &tree_iter, APROP_NAME, &tree_iter_name, -1);
                      if (strcmp (name, tree_iter_name) == 0)
                        {
                          gtk_list_store_remove (GTK_LIST_STORE(folder->storage), &tree_iter);
                          g_free (tree_iter_name);
                          break; 
                        }
                      g_free (tree_iter_name);
                      end_reached = gtk_tree_model_iter_next (folder->storage, &tree_iter);
                    }           
                  break;
                }
          }          
          g_free (cat);        
          g_free (name);
          gnome_desktop_file_free (p);
       }
       if (type)
           g_free (type);
  }
  else
     g_printerr ("Couldn't find \"%s\": %s\n", filename, err->message);

  if (err)
     g_error_free (err);
}

/*
 * Adds animated effects to the icons when being selected: enlarge the icon
 */
static void 
selection_iconview (GtkIconView *iconview, gpointer user_data)
{
  GtkTreeIter iter, former_iter;
  GtkTreePath *former_path = NULL; 
  gchar *icon, *former_icon;
  GdkPixbuf *ibuf, *former_ibuf;
  gchar *path_str, *former_path_str;
   
  GtkTreeModel *store = gtk_icon_view_get_model (iconview);
  GList *list = gtk_icon_view_get_selected_items (iconview);
  
  gint size_normal = get_item_size (ITEM_NORMAL);
  gint size_focused = get_item_size (ITEM_FOCUSED);
    
  /* reset the formerly selected item's icon to normal size */
  former_path_str = g_object_get_data (G_OBJECT (iconview), "former_selected");
  if (former_path_str)
      former_path = gtk_tree_path_new_from_string (former_path_str);
  if (former_path)
    {
      gtk_tree_model_get_iter (store, &former_iter, former_path);
      gtk_tree_model_get (store, &former_iter, APROP_ICONNAME, &former_icon, -1);
      former_ibuf = load_icon_scaled (former_icon, size_normal);  
      if (!former_ibuf)
        {
          former_ibuf = gpe_find_icon ("applauncher");
          former_ibuf = gdk_pixbuf_scale_simple (former_ibuf, size_normal, 
                                                 size_normal, GDK_INTERP_BILINEAR);	
        }
      if (former_ibuf)
        {
          gtk_list_store_set (GTK_LIST_STORE (store), &former_iter, 
                              APROP_ICONDATA, former_ibuf, -1);
          g_object_unref (former_ibuf);
        }
      g_free (former_path);
      g_free (former_icon);
   }
      
  if (g_list_length (list) > 0)
    {
      /* today is the yesterday of tomorrow */
      path_str = gtk_tree_path_to_string ((GtkTreePath*) list->data);
      g_object_set_data (G_OBJECT (iconview), "former_selected", path_str);		
  
      /* the selected item will have a slightly enlarged icon */	
      gtk_tree_model_get_iter (store, &iter, (GtkTreePath*) list->data);
      gtk_tree_model_get (store, &iter, APROP_ICONNAME, &icon, -1);	
      ibuf = load_icon_scaled (icon, size_focused);  
      if (!ibuf)
        {
          ibuf = gpe_find_icon ("applauncher");
          ibuf = gdk_pixbuf_scale_simple (ibuf, size_focused, size_focused, GDK_INTERP_BILINEAR);	
        }
      if (ibuf)
        {
          gtk_list_store_set (GTK_LIST_STORE (store), &iter, APROP_ICONDATA, ibuf, -1);
          g_object_unref (ibuf);
        }
      g_free (icon);
  }
    
  g_list_foreach (list, (GFunc)(gtk_tree_path_free), NULL);

  g_list_free (list);   
}

#ifdef GRIDLAYOUT
void 
set_super_folder_visible (void)
{
  GtkWidget *notebook = retrieve_notebook ();
	
  if (!notebook)
    return;
  gint pagenum = get_pagenum (_("Super"));
  gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), pagenum);
}

gboolean 
is_super_folder_visible (void)
{
  GtkWidget *notebook = retrieve_notebook ();
  if (!notebook)
    return FALSE;
  gint pagenum = get_pagenum (_("Super"));
  gint cur_pagenum = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
  if (cur_pagenum == pagenum)
    return TRUE;
  else
    return FALSE;
}
#endif

/* Create a new application tab from a given folder description. */
static void
new_application_tab (GtkWidget *notebook, t_folder *folder)
{
    GtkWidget *filebox, *sw;
    GtkTreeModel *store;
    
    store = GTK_TREE_MODEL (gtk_list_store_new (APROP_MAX, 
                                G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, 
                                G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, 
                                G_TYPE_BOOLEAN, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING));
    
    sw = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_NEVER, 
                                    GTK_POLICY_AUTOMATIC);
	
    filebox = gtk_icon_view_new_with_model (store);
    gtk_icon_view_set_columns (GTK_ICON_VIEW (filebox), COLUMNS_STACK);
    gtk_icon_view_set_spacing (GTK_ICON_VIEW (filebox), 0);
    gtk_icon_view_set_column_spacing (GTK_ICON_VIEW (filebox), gpe_get_boxspacing());
    gtk_icon_view_set_row_spacing (GTK_ICON_VIEW (filebox), gpe_get_boxspacing());
    gtk_icon_view_set_item_width (GTK_ICON_VIEW (filebox), -1);
	gtk_icon_view_set_markup_column (GTK_ICON_VIEW (filebox), APROP_NAME);
    gtk_icon_view_set_pixbuf_column (GTK_ICON_VIEW (filebox), APROP_ICONDATA);
    folder->iconview = filebox;	
	
    g_signal_connect (G_OBJECT(filebox), "item-activated", G_CALLBACK (applications_activate), folder);
	g_signal_connect (G_OBJECT(filebox), "selection-changed", G_CALLBACK (selection_iconview), NULL);
    
    folder->storage = gtk_icon_view_get_model (GTK_ICON_VIEW(filebox));
    
    /* packing */
    gtk_container_add (GTK_CONTAINER(sw), filebox);
    folder->pagenum = gtk_notebook_append_page (GTK_NOTEBOOK (notebook), sw, folder->icon);
}

static void 
settings_callback (int setting, int id, const char *key, type_t type, action_t action, void * user_data)
{  
  GList *iter;	
  gchar *file;
    
#ifdef DEBUG
  g_message ("applications callback for %s is called", key);
#endif
    
  setting_get_string (setting, key, &file);
  for (iter = folderlist; iter; iter = iter->next)
  {
    t_folder *folder = iter->data;
    render_bg (file, folder->iconview);   
  }
  g_free (file);  
}

/* Main function creating and setting up the application launcher user interface. */
void
applications_create (GtkWidget *notebook)
{
  GList *iter;
  gchar *str = NULL;
  gchar *current_theme = NULL; 
  gchar *bg_file = NULL;
  gboolean use_theme = FALSE;
  
  setting_get_string (setting, KEY_FOLDERS_PATH, &str);
	
  if (!str)
	{
      folderlist = load_folder_list (DEFAULT_FOLDERS_PATH);
	}
  else
    {
      folderlist = load_folder_list (str);
    }
	
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = (t_folder*) iter->data;         
      new_application_tab (notebook, folder);
    }
	 
#ifdef GRIDLAYOUT
  if (!str)
      load_application_list (DEFAULT_FOLDERS_PATH);
  else
    {
      load_application_list (str);
    }
#endif
  if (str)
    {
      g_free(str);
      str = NULL;
    }
	
  setting_get_string (setting, KEY_APPS_PATH, &str);
  if (!str)
	{
      load_application_list (DEFAULT_APPS_PATH);
	}
  else
    {
      load_application_list (str);
      g_free(str);
    }
	     	 
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      GtkTreePath *treepath = gtk_tree_path_new_from_string ("0");
	  gtk_icon_view_select_path (GTK_ICON_VIEW (folder->iconview), treepath);
	  gtk_tree_path_free (treepath);	 
    }	   
  applications_update_item_layout (LAYOUT_STACK);
  
  /*
   * The public setting can be changed outside of theme, e.g. by a image viewer,
   * so we have to take care of this setting  
   */
  int notify_id;
  setting_notify_add (setting, KEY_BG_IMAGE, settings_callback, NULL, &notify_id);
  
  /* 
   * Set background image, try public setting first, the background image 
   * from current theme comes second. 
   */
  
  setting_get_string (setting, KEY_CURRENT_THEME, &current_theme);
  if (current_theme == NULL)
    current_theme = g_strdup (DEFAULT_THEME);
    
  setting_get_string (setting, KEY_BG_IMAGE, &bg_file);
  if (bg_file == NULL)
    use_theme = TRUE;
  else if (access (bg_file, R_OK) != 0) //read access to the file is denied or this file doesn't exist
    use_theme = TRUE;  	
  
  if (use_theme)
    {
      gchar *key = g_strconcat (THEME_PREFIX, current_theme, KEY_THEME_BG, NULL);
      setting_get_string (setting, key, &bg_file);
      g_free (key);      
    }
  g_free (current_theme);
  
  if (!bg_file)
      bg_file = g_strdup (DEFAULT_THEME_BG);
  
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      render_bg (bg_file, folder->iconview);    
    }    
  g_free (bg_file);
  
  /* initialize startup notification */
  gpe_launch_install_filter ();
  gpe_launch_monitor_display (GDK_DISPLAY());
  
}
