/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * based on gpe-hello (c) 2005 Philippe De Swert, Florian Boor
 * 
 * (c) kernel concepts 2006, 2007
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Application launcher main, shared functions and initialisation.
 */

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>
#include <assert.h>
/* gtk includes */
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

/* i18n */
#include <libintl.h>
#include <locale.h> 

/* gpe includes */
#include <gpe/init.h>           /* gpe_application_init */
#include <gpe/pixmaps.h>        /* gpe_load_icons */
#include <gpe/spacing.h>        /* gpe_get_border, gpe_get_boxspacing */
#include <gpe/settings.h>
#include "gpe-applauncher.h"
#include "homescreen.h"
#include "applications.h"
#include "dbus-if.h"
#include "softkeys.h"
#include "hotkeys.h"
#include "launch-app.h"

#define _(x) gettext(x)

/* 
 * Forward declarations and
 * user interface controls.
 */
 
static void menu_new_message (void);
static void menu_new_contact (void);

/* User interface description for the GTK UI manager. */
static const gchar *ui_description = " \
    <ui>\
        <popup  action=\"PopupLeft\">\
            <menuitem name=\"message\" action=\"NewMessage\"/>\
            <menuitem name=\"contact\" action=\"NewContact\"/>\
        </popup>\
    </ui>\
"; 

static const GtkActionEntry action_entries[] = {
  {"NewMessage", NULL, "New _message", "", "Write a message", G_CALLBACK(menu_new_message)},
  {"NewContact", NULL, "New _contact", "", "Add a new contact", G_CALLBACK(menu_new_contact)},  
  {"PopupLeft", NULL, "_Popup", "", "Popup", NULL},
};

/*
 * This struct holds all pixmaps handled by libgpewidget functions.
 * It consists of a set of identifiers and icon files. For files in the
 * default GPE icon location you can omit the path and the extension.
 */
struct gpe_icon my_icons[] = 
{
  { "applauncher", "gpe-applauncher" },
  { "home", "stock_home" },
  { "loading" },
  { "msg_message" },
  { "msg_call" },
  { "msg_alarm" },	
  { NULL, NULL }
};

/*
 * Struct to hold some global data used in the application.
 */
struct
  {
    GtkWidget *mainwindow;
    GtkWidget *softkeywindow;
    GtkWidget *notebook;
    GtkWidget *maintabs;
    GtkWidget *menu_left;
    GdkGC *bg_gc;
    gboolean have_softkeys;
    gint  num_softkeys;
    guint *softkeys;
  } launcherui;

  
/* Handle theme change for pixmaps managed by libgpewidget. */
void 
set_gpe_theme (const gchar *theme)
{
   gpe_set_theme (theme);
   if (gpe_load_icons (my_icons) == FALSE)
     g_printerr ("Theme change failed, icons not updated.\n");
}

static gboolean 
popup_key_press_handler (GtkWidget *widget, GdkEventKey *event,
                         gpointer user_data)
{
  if (event->keyval == GDK_F4 || event->keyval == GDK_KP_F4)
    {
      const gchar *label_right = softkeys_get_label (1);
      if (label_right && strcmp (label_right, _("Cancel")) == 0)
          gtk_menu_popdown (GTK_MENU(launcherui.menu_left));
      softkeys_set_bg_color (1, PRESSED);    
      return TRUE;
    }
  return FALSE;
}

static gboolean 
popup_key_release_handler (GtkWidget *widget, GdkEventKey *event,
                           gpointer user_data)
{
  softkeys_set_bg_color (0, NORMAL);	
  return TRUE;
}

static gboolean 
popup_button_press_handler (GtkWidget *widget, GdkEventButton *event,
                           gpointer user_data)
{
  softkeys_set_bg_color (1, PRESSED);	
  return FALSE;
}

static gboolean 
popup_button_release_handler (GtkWidget *widget, GdkEventButton *event,
                           gpointer user_data)
{
  softkeys_set_bg_color (1, NORMAL);	
  return FALSE;
}

/* Create softkey menus with UI manager and description. */
static void
create_menus (void)
{
  GError *err = NULL;
  GtkActionGroup *action_group;
  GtkUIManager  *ui_manager;

  /* Set up action group and feed UI manager with actions and UI description. */
  action_group = gtk_action_group_new ("AllActions");
  ui_manager = gtk_ui_manager_new ();
    
  gtk_action_group_add_actions(action_group, action_entries, 
                               G_N_ELEMENTS(action_entries), NULL);
  gtk_ui_manager_insert_action_group (ui_manager, action_group, 0);
  if (!gtk_ui_manager_add_ui_from_string(ui_manager, ui_description, -1, &err))
    {
      g_printerr("Failed to parse menu data: %s\n", err->message);
      g_error_free(err);
      return;
    }
	
  launcherui.menu_left = gtk_ui_manager_get_widget(ui_manager, "/PopupLeft");	  
  g_signal_connect (launcherui.menu_left, "key-press-event", G_CALLBACK(popup_key_press_handler), NULL);	  
  g_signal_connect (launcherui.menu_left, "key-release-event", G_CALLBACK(popup_key_release_handler), NULL);
  g_signal_connect (launcherui.menu_left, "button-press-event", G_CALLBACK(popup_button_press_handler), NULL);
  g_signal_connect (launcherui.menu_left, "button-release-event", G_CALLBACK(popup_button_release_handler), NULL);
}

/* Menu callback for the "New message" action. */
static void 
menu_new_message (void)
{
    applications_start_service ("SMS", "send_sms", SERVICE_APPLAUNCHER);
}

/* Menu callback for the "New contact" action. */
static void 
menu_new_contact (void)
{
    applications_start_service ("ABook", "add_contact", SERVICE_APPLAUNCHER);
}

/* Switches from home screen to application launcher and back. */
void
applauncher_toggle_home (void)
{
  gint cur_tab = gtk_notebook_get_current_page (GTK_NOTEBOOK (launcherui.maintabs));
    
  gtk_notebook_set_current_page (GTK_NOTEBOOK (launcherui.maintabs), 
                                  cur_tab ? TAB_HOMESCREEN : TAB_APPLICATIONS);
}
  
/* Switches from mainmenu to home screen. */
static void
show_home (void)
{
  gtk_notebook_set_current_page (GTK_NOTEBOOK (launcherui.maintabs), 
                                 TAB_HOMESCREEN);
}

void 
lower_windows (Window root)
{
  static int level = 0;
  Window parent, *children;
  Display *display = GDK_DISPLAY();
  unsigned int nchildren;
  int stat, i;

  level++;

  stat = XQueryTree(GDK_DISPLAY(), root, &root, &parent, &children, &nchildren);
  if (stat == FALSE)
    {
      fprintf(stderr, "Can't query window tree...\n");
      return;
    }

  if (nchildren == 0)
      return;
  
  
  XLowerWindow (display, root);

  for(i=0; i < nchildren; i++)
   {
     XLowerWindow (display, children[i]);
     lower_windows(children[i]);
   }

  XFree((char *)children);
}

/* Handle softkey display if window is activated or gets the focus */
static gboolean 
window_activated (GtkWidget *window, GdkEventFocus *event, gpointer data)
{   	
  if (launcherui.num_softkeys == 2)
    {	  
      softkeys_update_key (0, _("Create"));
#ifdef GRIDLAYOUT
      if (gtk_notebook_get_current_page (GTK_NOTEBOOK(launcherui.maintabs)) == 0)
        softkeys_update_key (1, _("Toggle"));
      else
      {
        if (is_super_folder_visible ())
             softkeys_update_key (1, _("Toggle"));
        else
             softkeys_update_key (1, _("Back"));
      }
#else      
      softkeys_update_key (1, _("Toggle"));
#endif      
    }		

  return FALSE;
}


/* This menu positioning function determines the correct position for 
 * the softkey menus. */
static void
softkey_menu_position (GtkMenu *menu, gint *x, gint *y, gboolean *push_in,
                       gpointer user_data)
{
  gint nr = (gint) user_data;
  GtkRequisition req;
  gint pos_y;

  gtk_widget_size_request (GTK_WIDGET (menu), &req);

  pos_y = gdk_screen_height() - req.height - 24;
    
  if (launcherui.num_softkeys == 2)
    {
      if (nr == 0)
          *x = 1;
      else 
		if (nr == 1)
          *x = gdk_screen_width () - req.width - 1;
		
      *y = pos_y;
      *push_in = FALSE;
    } 
}

/* Activates a softkey menu. */
void
popup_softkey_menu (gint nr)
{  
  if (nr == 0)
    {
      if (!GTK_WIDGET_REALIZED (launcherui.menu_left))
        gtk_widget_realize (launcherui.menu_left);
      if (!GTK_WIDGET_MAPPED (launcherui.menu_left))
        gtk_widget_show (launcherui.menu_left);
            
      gtk_menu_popup (GTK_MENU (launcherui.menu_left), 
                      NULL, NULL, softkey_menu_position, (gpointer)nr, 1, 
                      gtk_get_current_event_time ());
	  gtk_menu_shell_select_first (GTK_MENU_SHELL (launcherui.menu_left), TRUE);
	  softkeys_update_key (1, _("Cancel"));	  	  
   }
  if (nr == 1)
    {	
#ifdef GRIDLAYOUT   	
      if (gtk_notebook_get_current_page (GTK_NOTEBOOK(launcherui.maintabs)) == 0)
        applauncher_toggle_home ();
      else
      {
        if (is_super_folder_visible ())
          applauncher_toggle_home ();
        else
          set_super_folder_visible ();
      }
#else
      applauncher_toggle_home ();
#endif      
    }
}
 
#ifdef GRIDLAYOUT 
void 
maintabs_switch_page (GtkNotebook *notebook, GtkNotebookPage *page,
                      guint page_num, gpointer user_data)
{
  if (page_num == 0)
    set_super_folder_visible ();
}  
                  
void 
notebook_switch_page (GtkNotebook *notebook,GtkNotebookPage *page,
					  guint page_num, gpointer user_data)
{
  /* When calling is_super_folder_visible () inside this callback, be noted that page_num
   * denotes the page which WILL become the 'current page' after this callback returns.
   * At this time, gtk_notebook_get_current_page returns the page which was the 'current page'.
   * So, if is_super_folder_visible () returns TRUE, this means super folder page is still the
   * 'current page' for this moment, although it is now not visible*/
  if (is_super_folder_visible ())
     softkeys_update_key (1, _("Back"));
}
#endif

static gboolean
window_key_release_event (GtkWidget *widget, GdkEventKey *k, gpointer data)
{
  gint i;
    
  for (i = 0; i < launcherui.num_softkeys; i++)
    {      
      if (k->keyval == launcherui.softkeys[i])
        {
          softkeys_set_bg_color (i, NORMAL);
#ifdef GRIDLAYOUT          
          if (is_super_folder_visible () && i == 1)
            softkeys_update_key (1, _("Toggle"));
#endif          	
          return TRUE;
        }
    }
  return FALSE;
}

/* global key handler for hot- and softkeys and digits */
static gboolean
window_key_press_event (GtkWidget *widget, GdkEventKey *k, gpointer data)
{
  gint i;
  
  /* softkeys */
  for (i = 0; i < launcherui.num_softkeys; i++)
    {      
      if (k->keyval == launcherui.softkeys[i])
        {
          softkeys_set_bg_color (i, PRESSED);	
          popup_softkey_menu (i);
          return TRUE;
        }
    }
  /* digits */
  if (k->keyval <= GDK_9 && k->keyval >= GDK_0)
    {
      if (gtk_notebook_get_current_page (GTK_NOTEBOOK(launcherui.maintabs)) == 0)
        {
          gchar *method = g_strdup_printf ("%d", (k->keyval - 48));
          applications_start_service (VOICECALL_APP_SERVICE, method, SERVICE_APPLAUNCHER);
          g_free (method);
          return TRUE;
        }
      else if (gtk_notebook_get_current_page (GTK_NOTEBOOK(launcherui.maintabs)) == 1)
       {
         gint digit = k->keyval - 49;
         gtk_notebook_set_current_page (GTK_NOTEBOOK(launcherui.notebook), digit);		  
       }
  }
	
  if ((k->keyval == GDK_Return) 
	  && (gtk_notebook_get_current_page (GTK_NOTEBOOK(launcherui.maintabs)) == 1))
    {
      gint page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (launcherui.notebook));
      GtkWidget *child = gtk_notebook_get_nth_page (GTK_NOTEBOOK (launcherui.notebook), page_num);		  												
      GtkWidget *iconview = gtk_bin_get_child (GTK_BIN (child));
      GList *list = gtk_icon_view_get_selected_items (GTK_ICON_VIEW (iconview));
      if (g_list_length (list) > 0)
        {
          gtk_icon_view_item_activated (GTK_ICON_VIEW (iconview), list->data);
          g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
          g_list_free (list);
        }
      return TRUE;
    }
 
#ifdef DEBUG
  g_message ("%s: pass on the key press event", __FUNCTION__);
#endif
  return FALSE;
}


static gboolean
client_event (GtkWidget *widget, GdkEventClient *event, gpointer user_data)
{
	if (event->data.l[0] == 100)// this event comes from MB right after 'show desktop'
      {
		show_home ();
#ifdef DEBUG
		g_message ("show home which is the desktop");
#endif
      }
	return TRUE;
}
	
/* Load a named icon and scale it to the designated size. */
GdkPixbuf*
load_icon_scaled (const gchar *iconname, gint size)
{
  GdkPixbuf *pbuf, *result;
  gchar *str = NULL, *fullname = NULL;   
  gchar *current_theme = NULL;  
  gchar *resource_path = NULL;
	
  if (iconname && strstr(iconname, ".png"))
      fullname = g_strdup_printf ("%s", iconname);
  else
      fullname = g_strdup_printf ("%s.png", iconname);

  if (fullname[0] == G_DIR_SEPARATOR)
    {
      str = g_strdup (fullname);
    }
  else  
    {
      setting_get_string (setting, KEY_CURRENT_THEME, &current_theme);
      if (current_theme == NULL)
        current_theme = g_strdup ("default");
        
      setting_get_string (setting, KEY_RESOURCE_PATH, &resource_path);
      if (resource_path == NULL)
          resource_path = g_strdup (DEFAULT_RESOURCE_PATH);
      
      str = g_strconcat (resource_path, current_theme, "/", fullname, NULL);  
      g_free (current_theme);
      g_free (resource_path);  
      
      if (g_file_test (str, G_FILE_TEST_EXISTS) == FALSE)
        {
          g_free (str);
          str = g_strconcat (DEFAULT_PIXMAP_PATH, "/", fullname, NULL);  
        }
    }
    
  if (g_file_test (str, G_FILE_TEST_EXISTS) == FALSE)
    {
      pbuf = NULL;
      result = NULL;
    }
  else
    {	  
      pbuf = gdk_pixbuf_new_from_file (str, NULL);	  
      if (pbuf != NULL)      	  
        result = gdk_pixbuf_scale_simple (pbuf, size, size, GDK_INTERP_BILINEAR);      
      else
        result = NULL;
    }
  
  if (pbuf) 
    g_object_unref (pbuf);

#ifdef DEBUG
  g_message ("load pixmap: %s", str);
#endif
  
  g_free (str);
  g_free (fullname);
  
  return result;
}

/* Set up softkey usage and initialize label texts. */
gboolean
setup_softkeys (void)
{
  gint i;
  const gchar *key;
    
  launcherui.num_softkeys = softkeys_get_amount ();
  launcherui.softkeys = g_malloc0 (launcherui.num_softkeys * sizeof (guint));

  for (i = 0; i < launcherui.num_softkeys; i++)
    {      
      key = softkeys_get_key (i);
      if (key)
        {
          launcherui.softkeys[i] = gdk_keyval_from_name (key);
        }
    }
        
  /* this is currently the only handled softkey configuration */
  if (launcherui.num_softkeys == 2)
    {      
      softkeys_update_key (0, _("Create"));
      softkeys_update_key (1, _("Toggle"));
      return TRUE;
	}
  return FALSE;  
}

/* Present the application launcher and show the given tab. */
void
applauncher_present (gint tab)
{
  if ((tab == TAB_HOMESCREEN) || (tab == TAB_APPLICATIONS))
    gtk_notebook_set_current_page (GTK_NOTEBOOK (launcherui.maintabs), tab);
  gtk_window_present (GTK_WINDOW (launcherui.mainwindow));
}

/* Get the internal softkey area widget. */
GtkWidget *
applauncher_get_softkey_area (void)
{
  return launcherui.softkeywindow;
}

/* Get the internal main window widget. */
GtkWidget *
applauncher_get_main_window (void)
{
  return launcherui.mainwindow;
}

/* Changes the background colour of a widget supporting this. */
void
set_background_colour (GdkColor colour, GtkWidget *widget)
{
  GdkColormap *map;
    
  map = gdk_drawable_get_colormap (GTK_WIDGET (launcherui.mainwindow)->window);
  gdk_colormap_alloc_color (map, &colour, FALSE, TRUE);
  gtk_widget_modify_bg (widget, GTK_STATE_NORMAL, &colour);
}

void
render_bg (const gchar* filename, GtkWidget *widget)
{
  GError *error = NULL;
  GdkPixbuf *gdkpixbuf;
  GdkPixmap *bgPixmap;
  GdkBitmap *bgMask;

#ifdef DEBUG    
  g_print ("%s: bg image file is %s\n", __FUNCTION__, filename);
#endif
    
  gdkpixbuf = gdk_pixbuf_new_from_file (filename, &error);
  if (error != NULL)
    {
      g_printerr ("<%s> unable to find background file: %s\n", __FUNCTION__, error->message);
      g_error_free (error);
    }
  gdk_pixbuf_render_pixmap_and_mask (gdkpixbuf, &bgPixmap, &bgMask, 1);
  if (bgPixmap == NULL)
    {
      g_printerr ("<%s> no pixmap created!\n", __FUNCTION__);
    }
   
  GtkStyle *style = gtk_style_copy (widget->style);
   
  if (style->bg_pixmap[0])
    g_object_unref (style->bg_pixmap[0]);
  style->bg_pixmap[0] = g_object_ref (bgPixmap);
  gtk_widget_set_style (widget, style);
  g_object_unref (style);
 
//  GtkStyle * gtkStyle;
//  gtkStyle = gtk_rc_get_style (widget);
//  gtkStyle->bg_pixmap[0] = bgPixmap;
//  gtk_style_set_background (gtkStyle, widget->window, GTK_STATE_NORMAL);
}

/* Set a background image to a widget supporting backgrounds. */
void
set_background_pixmap (const gchar *filename, GtkWidget *widget)
{
  GdkPixbuf *pbuf, *bg;
  GdkPixmap *pixmap;
  gint height, width;

#ifdef DEBUG    
  g_print ("%s: bg image file is %s\n", __FUNCTION__, filename);
#endif
    
  if (!GTK_WIDGET_REALIZED (widget))
    gtk_widget_realize (widget);

  if (filename == NULL)
    {
      gdk_window_set_back_pixmap(widget->window, NULL, FALSE);
      return;
    }
    
  pbuf = gdk_pixbuf_new_from_file (filename, NULL); /* silently ignore error */
  
  if (pbuf == NULL) 
      return;
    	
  /* determine dimensions */
  width = gdk_pixbuf_get_width(pbuf);
  height = gdk_pixbuf_get_height(pbuf);
	
  /* compose a new pixbuf with background instead of alpha channel to avoid 
     making random fb background visible. */
  bg = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 
                      gdk_pixbuf_get_bits_per_sample(pbuf), width, height);
  gdk_pixbuf_fill(bg, 0);
  gdk_pixbuf_composite(pbuf, bg, 0, 0, width, height, 0, 0, 1, 1, 
                       GDK_INTERP_BILINEAR, 0xFF);
  
  pixmap = gdk_pixmap_new(widget->window,  width, height, -1);
  gdk_draw_pixbuf(pixmap, NULL, bg, 0, 0 , 0, 0, -1, -1, 
	              GDK_RGB_DITHER_NORMAL, 1, 1);
  g_object_unref(pbuf);
  g_object_unref(bg);

  gtk_widget_set_app_paintable (widget, TRUE);
  gdk_window_set_back_pixmap(widget->window, pixmap, FALSE);
  
  gtk_widget_queue_draw (widget);
}

/* Command line help display function. */
static void
show_help (void)
{
  printf("\nGPE Application Launcher\n\n");
  printf("Usage: gpe-applauncher [-d|-h]\n");
  printf("  -d | --no-desktop: Do not become a desktop window.\n");
  printf("  -h | --help: Display this help.\n");
}


static gboolean
start_dialer (gpointer data)
{
#ifdef DEBUG
  g_message ("into start_dialer");
#endif
  applications_start_service (VOICECALL_APP_SERVICE, "run_and_hide", 
                              SERVICE_APPLAUNCHER);

  return FALSE;
}


/* Main routine, application startup, initialisation and creating of all main 
 * components such as home screen and application launcher. */
int
main (int argc, char *argv[])
{
  GtkWidget *notebook, *vbox, *space_panel, *space_softkeys, *maintabs;
  gboolean flag_nodesktop = FALSE;
  gint width = 300;
  gint height = 400;
  gint ypos = 0;

  if (gpe_application_init (&argc, &argv) == FALSE)
    exit (1);
  
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  bind_textdomain_codeset (PACKAGE, "UTF-8");
  textdomain (PACKAGE);

  /* parse command line options */
  for (;;)
    {
      int option_index = 0;
      int c;

      static struct option long_options[] = {
		{"no-desktop", 0, 0, 'd'},
		{"help", 0, 0, 'h'},
		{0, 0, 0, 0}
      };
      
      c = getopt_long (argc, argv, "dh", long_options, &option_index);
      if (c < 0)
        break;

      switch (c)
        {
        case 'd':
          flag_nodesktop = TRUE;
          break;
        case 'h':
          show_help ();
          exit (0);
          break;
        }
    }

  /* open the connection with setting system */
  setting = setting_open ();	
  
  char * current_theme = NULL;  
  setting_get_string (setting, KEY_CURRENT_THEME, &current_theme);
  if (current_theme == NULL)
    current_theme = g_strdup ("default");
    
  set_gpe_theme (current_theme);
  if (gpe_load_icons (my_icons) == FALSE)
    exit (1);	
  
  /* create application window */
  launcherui.mainwindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  gtk_widget_add_events (launcherui.mainwindow, GDK_KEY_PRESS_MASK);
  
  gtk_widget_realize (launcherui.mainwindow);
 
  if (flag_nodesktop == FALSE)
    {	  
      gdk_window_set_type_hint (launcherui.mainwindow->window, GDK_WINDOW_TYPE_HINT_DESKTOP);
      width = gdk_screen_width ();
      height = gdk_screen_height ();
      ypos = (height > 400) ? PANEL_LARGE : PANEL_SMALL; /* softkey area */
    }
    
  gtk_window_set_default_size (GTK_WINDOW (launcherui.mainwindow), width, height);
  
  	
  g_signal_connect (G_OBJECT (launcherui.mainwindow), "delete-event", gtk_main_quit, NULL);
  gtk_window_set_title (GTK_WINDOW (launcherui.mainwindow), "Application Launcher");
  
  /* show icon in window decoration */
  gpe_set_window_icon (launcherui.mainwindow, "applauncher");
	
 
  maintabs = gtk_notebook_new();
  launcherui.maintabs = maintabs;
  gtk_notebook_set_scrollable(GTK_NOTEBOOK(maintabs), FALSE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK(maintabs), FALSE);
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK(maintabs), FALSE);
  GTK_WIDGET_UNSET_FLAGS (maintabs, GTK_CAN_FOCUS);

#ifdef GRIDLAYOUT
  g_signal_connect (G_OBJECT (launcherui.maintabs), "switch-page", 
                    G_CALLBACK (maintabs_switch_page), NULL);
#endif
  	
  notebook = gtk_notebook_new();
  launcherui.notebook = notebook;
  gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook), TRUE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK(notebook), FALSE);
  
#ifdef GRIDLAYOUT
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK(notebook), FALSE);
  g_signal_connect (G_OBJECT (launcherui.notebook), "switch-page", G_CALLBACK (notebook_switch_page), NULL);
#endif
  
  vbox = gtk_vbox_new(FALSE, 0);
  space_panel = gtk_label_new(NULL);
  gtk_label_set_label (GTK_LABEL (space_panel), "panel");
  gtk_widget_set_size_request (space_panel, -1, ypos);
  space_softkeys = gtk_label_new(NULL);
  gtk_label_set_label (GTK_LABEL (space_softkeys), "softkey"); 
  gtk_widget_set_size_request (space_softkeys, -1, ypos);

  /* homescreen */
  homescreen_create (maintabs);
  
  /* now fill the notebook with tabs */
  applications_create (notebook);
  
  gtk_notebook_append_page (GTK_NOTEBOOK (maintabs), notebook, NULL); 
	
  /* make everything visible and start the main loop */
  gtk_box_pack_start (GTK_BOX(vbox), space_panel, FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX(vbox), maintabs, TRUE, TRUE, 0);  
  gtk_box_pack_start (GTK_BOX(vbox), space_softkeys, FALSE, TRUE, 0);
  gtk_container_add (GTK_CONTAINER (launcherui.mainwindow), vbox);  


  gtk_widget_show_all (launcherui.mainwindow);  

  /* initialize settings and bus interfaces */
  launcherui.bg_gc = NULL;

  dbus_init ();

  /* install global key handler */ 
  g_signal_connect (G_OBJECT (launcherui.mainwindow), "key_press_event", 
                    G_CALLBACK (window_key_press_event), NULL);
  g_signal_connect (G_OBJECT (launcherui.mainwindow), "key_release_event", 
                    G_CALLBACK (window_key_release_event), NULL);
                    					
  g_signal_connect (G_OBJECT (launcherui.mainwindow), "client_event", 
                    G_CALLBACK (client_event), NULL);					

  /* create softkey area */
  launcherui.softkeywindow = softkeys_create_window();
  if (launcherui.softkeywindow) 
    {
      softkeys_update_theme (current_theme);
      gtk_widget_show_all (launcherui.softkeywindow);
    }
  g_free (current_theme);	
	
  /* set up softkey usage */
  launcherui.have_softkeys = setup_softkeys ();
  if (launcherui.have_softkeys)
    {
      g_signal_connect (G_OBJECT (launcherui.mainwindow), "focus-in-event", 
                        G_CALLBACK (window_activated), NULL);
        
      /* create softkey menus */
      create_menus ();
    }

  gtk_widget_grab_focus (launcherui.notebook);
	
  /* make sure we keep the process table clean */
  signal (SIGCHLD, SIG_IGN);
  		
  /*
   * Being an important application of phone, dialer needs to be started
   * in rapid sequence
   */		
  g_timeout_add (1500, start_dialer, NULL);


  /*
   * We need to direct certain key/mouse events to certain windows which may 
   * not be able to receive these events normally since they happen on 
   * applauncher's softkey area. 
   */
  hotkeys_initialize (launcherui.mainwindow);

  if (launcherui.softkeywindow) 
      gtk_window_present (GTK_WINDOW (launcherui.softkeywindow));

  launch_init ();
   
  gtk_main ();

  exit (0);
}
