/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * 
 * (c) kernel concepts 2006, 2007
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Application launcher user interface.
 */
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <libintl.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <gpe/spacing.h>
#include <gpe/infoprint.h>
#include <gpe/launch.h>
#include <gpe/desktop_file.h>
#include <gpe/pixmaps.h>
#include <gpe/gtkappnotification.h>
#include <gpe/settings.h>

#include "gpe-applauncher.h"
#include "launch-app.h"
#include "dbus-if.h"

#define _(x) gettext (x)

#define ICON_LOADING "loading.gif"
#define VOC_PID "/tmp/.voc.pid"

/* to test whether a lock file can be acquired with Write permission
   return 0 means OK
   return 1 means another process is having lock on this file

   The Voc application will create a lock file when started, the file will 
   exist in the system when Voc is running until Voc exits. Applauncher tests
   the existence of this file to determine whether Voc is running (return 1)
   or not (return 0)
*/
static int 
test_lock(const gchar * lock_file)
{
  int fd, ret;
  struct flock lock_it;

  fd = open (lock_file, O_WRONLY);
  
  lock_it.l_type = F_WRLCK;
  lock_it.l_whence = SEEK_SET;
  lock_it.l_start = 0;
  lock_it.l_len = 0;

  if (fcntl (fd, F_GETLK, &lock_it) < 0)
      ret = 0;
  else if (lock_it.l_type == F_UNLCK)
      ret = 0;
  else 
      ret = 1;
  
  close (fd);
  return ret;
}

static GHashTable * notification_table;

static gboolean 
launch_notification_timeout_cb (gpointer callee)
{
  GtkWidget * nc_window = NULL;
            
  nc_window = (GtkWidget *)g_hash_table_lookup(notification_table, callee);
  if (nc_window)
    {
      gtk_widget_destroy(nc_window);
      g_hash_table_remove(notification_table, callee);
    }

  g_free(callee);

  return FALSE;
}

/* Callback used to track the startup status of applications. */
static void
launch_status_callback (enum launch_status status, void *data)
{
  gchar *callee = NULL;
    
  if (data)
    {
      gchar ** vec = g_strsplit_set((gchar*)data, "!", -1);
		
      if (vec[0])
        {
          callee = g_strdup(vec[0]);
#ifdef DEBUG            
          g_print("%s(): callee '%s'", __FUNCTION__, callee ? callee : "null");
#endif            
        }
		
        g_strfreev(vec);
    }
   
  switch (status)
    {
      case LAUNCH_STARTING:
#ifdef DEBUG
        g_message ("Launch starting...");
#endif        
        if (callee)
          {
            GtkWidget *nc_window = NULL;
            gchar *file;
            gchar *resource_path = NULL;
            gchar *current_theme = NULL;

            setting_get_string (setting, KEY_CURRENT_THEME, &current_theme);
            if (current_theme == NULL)
              current_theme = g_strdup (DEFAULT_THEME);
            setting_get_string (setting, KEY_RESOURCE_PATH, &resource_path);
            if (!resource_path)
                resource_path = g_strdup (DEFAULT_RESOURCE_PATH);
            file = g_strconcat (resource_path, current_theme, "/", ICON_LOADING, NULL);
            g_free (current_theme);         
            g_free (resource_path);
             
            nc_window = gtk_app_notification_new();
            gtk_widget_set_size_request(GTK_WIDGET(nc_window), -1, 70);
            gtk_app_notification_set_text(GTK_APP_NOTIFICATION(nc_window), _("Launching..."));
    
            gtk_app_notification_set_image_from_file(GTK_APP_NOTIFICATION(nc_window), file);
            g_free (file);
            
            gtk_app_notification_hide_progressbar(GTK_APP_NOTIFICATION(nc_window));
            
            gtk_window_set_position(GTK_WINDOW(nc_window), GTK_WIN_POS_CENTER_ALWAYS);

            g_hash_table_replace(notification_table, (gpointer)callee, (gpointer)nc_window);
                                                     
            g_timeout_add(4 * 1000, launch_notification_timeout_cb, g_strdup(callee));
                                                     
            gtk_widget_show_all(nc_window);
          }
        
        break;
      
      case LAUNCH_FAILED:      		
#ifdef DEBUG
        g_message ("Launch failed!");
#endif        
      case LAUNCH_COMPLETE:
#ifdef DEBUG
        g_message ("Launch completed!");
#endif        
        if (data)
          {
            send_message (data);		
            g_free (data);
          }
        if (callee)
          {
            GtkWidget *nc_window = NULL;
            
            nc_window = (GtkWidget *)g_hash_table_lookup(notification_table, (gpointer)callee);
            if (nc_window)
              {
                gtk_widget_destroy(nc_window);
                g_hash_table_remove(notification_table, (gpointer)callee);
              }
          }
        break;        
    }
}

gboolean 
launch_application (const char *name, const char *executable, const char *argv,
                    gboolean single_instance, gboolean startup_notify,
					const char *caller)
{
  Window w;
  gchar *s;
  gboolean ret = FALSE;
    
  if (!name || !executable || !argv || !caller)
      return FALSE;
  
  gchar *exec = g_strdup (executable);
    
  if (!startup_notify)
      single_instance = FALSE;
    
  s = strstr (exec, "%U");
  if (!s)
      s = strstr (exec, "%f");
  if (s)
      *s = 0;
                      
  gchar *exec_argv = g_strconcat (exec, " ", argv, NULL);
  gchar *str_send = g_strdup_printf ("%s!%s!%s", name, argv, caller);
  
#ifdef DEBUG
  g_message ("name = '%s'", name);
  g_message ("exec = '%s'", exec);
  g_message ("exec_argv = '%s'", exec_argv);
  g_message ("caller = '%s'", caller);
#endif
  
  if (!single_instance)
    {
      /* run another instance */  											  
      ret = gpe_launch_program_with_callback (GDK_DISPLAY(), exec_argv, name, 
                                              startup_notify,
                                              launch_status_callback, str_send);
    }
  else
    { 
      /* Voc will place a file in /tmp to advertise its existence */
      if (strcmp (name, "Voc") == 0) 
        {
          if (test_lock (VOC_PID))
            {
              send_message (str_send);	
              g_free (str_send);
              ret = TRUE;
            }
          else
            {			
              ret = gpe_launch_program_with_callback (GDK_DISPLAY(), exec_argv, 
                                                      name, TRUE,
                                                      launch_status_callback, str_send);
            }
        }
      else /* the running-or-not determination of the other applications are 
           conducted by checking with its window */
        {
            w = gpe_launch_get_window_for_binary (GDK_DISPLAY(), exec);  		  
            if (w)
              {					  
                /* tell the caller that startup is successful */  
#ifdef DEBUG
                g_message ("activate it");  
#endif
                if (strcmp (argv, "") == 0)
                    gpe_launch_activate_window (GDK_DISPLAY(), w);
                else
                  {
                    send_message (str_send);	
                    g_free (str_send);
                  }
                ret = TRUE;						
              }
            else
              {
                if (!gpe_launch_startup_is_pending (GDK_DISPLAY(), exec))
                  {
                    /* Actually run the program */
#ifdef DEBUG
                    g_message ("launch it");
#endif
                    ret = gpe_launch_program_with_callback (GDK_DISPLAY(), 
                                            exec_argv, name, TRUE,
                                            launch_status_callback, str_send);
                  }
              } 
        }
    }
  g_free (exec_argv);
  g_free (exec);
  return ret;
}

void launch_init (void)
{
  notification_table = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
}

