/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * 
 * (c) kernel concepts 2006, 2007
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * Softkey window and management implementation.
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/* X includes */
#include <X11/Xlib.h>
#include <X11/Xatom.h>

/* gtk includes */
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>

/* i8n */
#include <libintl.h>
#include <locale.h> 

/* gpe includes */
#include <gpe/spacing.h> 
#include <gpe/settings.h>

#include "gpe-applauncher.h"
#include "settings.h"
#include "softkeys.h"

#define _(x) gettext(x)

/*setting key definition*/
#define KEY_THEME_BG "/softkeyarea/backgroundimage/name"

/*file location definition*/
#define DEFAULT_THEME_BG PREFIX "/share/themes/default/applauncher_softkey_area.png" 
#define FILE_CONFIG     SYSCONFDIR "/gpe/softkeys.conf"

static struct
{
  t_softkey *softkeys;
  gint num_softkeys;
  gchar *label_colour;
  gint bar_height;
} skbar;

/* Returns the total amount of defined softkeys. */
gint 
softkeys_get_amount (void)
{
  return skbar.num_softkeys;
}

/* Returns the key binding of a softkey definition. */
const gchar*
softkeys_get_key (gint nr)
{
  if ((nr >= 0) && (nr < skbar.num_softkeys))
    return skbar.softkeys[nr].key;
  else
    return NULL;
}

/* Returns the label of a softkey. */
const gchar*
softkeys_get_label (gint nr)
{
  if ((nr >= 0) && (nr < skbar.num_softkeys))
    return skbar.softkeys[nr].label;
  else
    return NULL;
}
	
/* Returns the softkey menu X and Y position for the selected key. */
gboolean 
softkeys_get_menu_pos (gint nr, gint *menu_x, gint *menu_y)
{
  if ((nr >= 0) && (nr < skbar.num_softkeys))
    {
      *menu_x = skbar.softkeys[nr].menu_x;
      *menu_y = skbar.softkeys[nr].menu_y + skbar.bar_height;
      return TRUE;
    }
  else
    {
      *menu_x = 0;
      *menu_y = 0;
      return FALSE;
    }
}

/* Changes the text on a given softkey. */
void
softkeys_update_key (gint nr, const gchar *labeltext)
{
  if (nr < skbar.num_softkeys)
    {
      if (skbar.softkeys[nr].widget)
        {
          gchar *text;
          GtkWidget *lwidget;              
          lwidget = gtk_bin_get_child(GTK_BIN(skbar.softkeys[nr].widget));
			
		  if (skbar.label_colour == NULL)
            {
              gtk_label_set_text (GTK_LABEL (lwidget), labeltext);			  
            }
          else
            {              
              text = g_markup_printf_escaped ("<span weight=\"bold\" foreground=\"%s\">%s</span>", 
                                             skbar.label_colour, labeltext);
              gtk_label_set_markup (GTK_LABEL (lwidget), text);
              g_free (text);
            }
        }
      if (skbar.softkeys[nr].label)
          g_free (skbar.softkeys[nr].label);
      skbar.softkeys[nr].label = g_strdup (labeltext);
    }
}

/* Refreshes all softkeys to the applications own text and current colour. */
static void
softkeys_refresh_labels (void)
{
  gint i;
    
  for ( i = 0; i < skbar.num_softkeys; i++)
    {
	  gchar *text;
      GtkWidget *lwidget;
      lwidget = gtk_bin_get_child (GTK_BIN(skbar.softkeys[i].widget));
		
      if (skbar.label_colour == NULL)
        {
          gtk_label_set_text (GTK_LABEL (lwidget), skbar.softkeys[i].label);
        }
      else
        {          
          text = g_markup_printf_escaped ("<span weight=\"bold\" foreground=\"%s\">%s</span>", 
                                         skbar.label_colour, skbar.softkeys[i].label);
          gtk_label_set_markup (GTK_LABEL (lwidget), text);
          g_free (text);
        }
    }
}


/* Change the colour used for softkey labels. */
void
softkeys_update_label_colour (const gchar *colour)
{
  if (skbar.label_colour)
    g_free (skbar.label_colour);
  skbar.label_colour = g_strdup (colour);
  softkeys_refresh_labels ();
}

void 
softkeys_set_bg_color (gint nr, State state)
{
  gchar *text;
  GtkWidget *lwidget;
	
  lwidget = gtk_bin_get_child (GTK_BIN(skbar.softkeys[nr].widget));
  if (state == PRESSED)
	text = g_markup_printf_escaped ("<span weight=\"bold\" background=\"%s\">%s</span>", 
                                         "orange", skbar.softkeys[nr].label);
  else
    text = g_markup_printf_escaped ("<span weight=\"bold\" foreground=\"%s\">%s</span>", 
                                    skbar.label_colour, skbar.softkeys[nr].label);                                         
  gtk_label_set_markup (GTK_LABEL (lwidget), text);
  g_free (text);                                     
}

/* Load the softkey definitons from configuration file. Returns FALSE
 * If no keys defined or an error occured. */
static gboolean
softkeys_load_configuration (void)
{
  GKeyFile *configfile;
  GError *err = NULL;
  gchar **keydefs, *akeydef;

  configfile = g_key_file_new();
  		
  if (!g_key_file_load_from_file (configfile, FILE_CONFIG,
                                  G_KEY_FILE_KEEP_COMMENTS, &err))
	{
      fprintf(stderr, _("No softkeys defined - disabling."));
	  g_error_free(err);
	  err = NULL;
      return FALSE;
	}
    
  keydefs = g_key_file_get_groups (configfile, NULL);
  if (keydefs == NULL)
      return FALSE;
  
  while (keydefs[skbar.num_softkeys])
    {
      gboolean error_occured = FALSE;
        
      skbar.softkeys = g_realloc (skbar.softkeys, sizeof (t_softkey) * (skbar.num_softkeys + 1));
      memset (&skbar.softkeys[skbar.num_softkeys], 0, sizeof (t_softkey));
      akeydef = keydefs[skbar.num_softkeys];
        
      skbar.softkeys[skbar.num_softkeys].label = g_key_file_get_string(configfile, akeydef, "label", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].width = g_key_file_get_integer(configfile, akeydef, "width", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].height = g_key_file_get_integer(configfile, akeydef, "height", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].x = g_key_file_get_integer(configfile, akeydef, "xpos", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].y = g_key_file_get_integer(configfile, akeydef, "ypos", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].menu_x = g_key_file_get_integer(configfile, akeydef, "xpos_menu", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].menu_y = g_key_file_get_integer(configfile, akeydef, "ypos_menu", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].alignment = (gfloat)g_key_file_get_integer(configfile, akeydef, "align", &err) / 10;
      if (err != NULL)
        {
            skbar.softkeys[skbar.num_softkeys].alignment = 0.5;
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      skbar.softkeys[skbar.num_softkeys].key = g_key_file_get_string(configfile, akeydef, "key", &err);
      if (err != NULL)
        {
            error_occured = TRUE;
            g_error_free (err);
            err = NULL;
        }
      if (error_occured == FALSE)
        skbar.num_softkeys++;
    }
    
  g_strfreev(keydefs);
  return (skbar.softkeys ? TRUE : FALSE);
}

static XKeyEvent
create_x_key_event(Display *display, Window window, Window window_root, gboolean down, guint keycode)
{
  static XKeyEvent event;
  event.display     = display;
  event.window      = window;
  event.root        = window_root;
  event.subwindow   = None;
  event.state       = 0;
  event.x           = 1;
  event.y           = 1;
  event.x_root      = 1;
  event.y_root      = 1;
  event.same_screen = True;
  event.keycode     = XKeysymToKeycode(display, keycode);
  event.time        = CurrentTime;
    
  if(down)
    event.type = KeyPress;
  else
    event.type = KeyRelease;
    
  return event;
}

static Window
get_active_window (void)
{
  Atom realtype;
  int format;
  int status;
  Window ret;
  unsigned long n;
  unsigned long extra;
  union
    {
      Window *win;
      unsigned char *char_pointer;
    } win;
 
  Atom active_app_atom =
          XInternAtom (GDK_DISPLAY (), "_MB_CURRENT_APP_WINDOW", False);

  win.win = NULL;

  status = XGetWindowProperty (GDK_DISPLAY(), GDK_ROOT_WINDOW(),
                               active_app_atom, 0L, 16L,
                               0, XA_WINDOW, &realtype, &format,
                               &n, &extra, &win.char_pointer);
  if (!(status == Success && realtype == XA_WINDOW && format == 32
                && n == 1 && win.win != NULL))
    {
      if (win.win != NULL)
          XFree (win.char_pointer);
      return GDK_FILTER_CONTINUE;
    }

  ret = win.win[0];

  if (win.win != NULL)
      XFree(win.char_pointer);
    
  return ret;
}

static gboolean
on_softkey_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  gint nr = (gint)user_data;
	
  softkeys_set_bg_color (nr, PRESSED);
  return FALSE;
}

static gboolean
on_softkey_released (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  gint nr = (gint)user_data;
  softkeys_set_bg_color (nr, NORMAL);
    
  Window toplevel_window;
  int    revert = RevertToNone;
  guint keysym = gdk_keyval_from_name (softkeys_get_key (nr));
  XKeyEvent keyevent;
  GtkWidget * mainwindow = applauncher_get_main_window();
    
  XGetInputFocus(GDK_DISPLAY(), &toplevel_window, &revert);   

  if ( GDK_WINDOW_XID (mainwindow->window) == toplevel_window)
    {
      popup_softkey_menu (nr);
    }
  else
    {
      keyevent = create_x_key_event(GDK_DISPLAY(), toplevel_window, 
                           GDK_ROOT_WINDOW(), TRUE, keysym);
      XSendEvent(keyevent.display, keyevent.window, True, KeyPressMask, 
                 (XEvent*)&keyevent);
    }
  return TRUE;

}

/* Add a softkey widget to a given layout. */
static void
add_softkey (t_softkey *skey, GtkFixed *parent, gint nr)
{  
  GtkWidget *label;
  GtkWidget *eventbox;
	
  label = gtk_label_new (skey->label);	
  eventbox = gtk_event_box_new ();


  gtk_widget_set_size_request (eventbox, skey->width, skey->height);
  gtk_event_box_set_visible_window (GTK_EVENT_BOX (eventbox), FALSE);
  gtk_container_add (GTK_CONTAINER (eventbox), label);	
  gtk_misc_set_alignment (GTK_MISC (label), nr, 1);
  gtk_misc_set_padding (GTK_MISC (label), 2, 2);	
  
  gtk_fixed_put (parent, eventbox, skey->x, skey->y);
	
  skey->widget = eventbox;   
  					 
  gtk_widget_add_events (GTK_WIDGET (eventbox), GDK_ALL_EVENTS_MASK);
  g_signal_connect (G_OBJECT (eventbox), "button-press-event",
                     G_CALLBACK(on_softkey_pressed), (gpointer)nr);
  g_signal_connect (G_OBJECT (eventbox), "button-release-event",
                     G_CALLBACK(on_softkey_released), (gpointer)nr);                     						 
}

/* Callback function for theme change. Reads settings from the new theme. */
void 
softkeys_update_theme (const gchar* themename)
{
  gchar *bg_file = NULL;
  gchar *key = g_strconcat (THEME_PREFIX, themename, KEY_THEME_BG, NULL);
  GtkWidget *softkeywindow = applauncher_get_softkey_area ();
    
  setting_get_string (setting, key, &bg_file);
  g_free (key);
  if (bg_file)
    {
      set_background_pixmap (bg_file, softkeywindow);    
      g_free (bg_file);
    }
  else
      set_background_pixmap (DEFAULT_THEME_BG, softkeywindow);
}

/* Function to build up and initilize the softkey area. */
GtkWidget *
softkeys_create_window (void)
{
  gint width, height;
  GtkWidget *window;
  GdkGeometry geometry;
  GtkWidget *softkey_area;
  gint i;
  gchar *key = NULL;
  gchar *current_theme = NULL; 
  gchar *bg_file = NULL;
    
  skbar.num_softkeys = 0;
  skbar.label_colour = g_strdup("white");
  skbar.softkeys = NULL;
  
  if (softkeys_load_configuration () == FALSE) 
    return NULL;
  
  width = gdk_screen_width ();
  height = (gdk_screen_height () > 400) ? PANEL_LARGE : PANEL_SMALL;
  geometry.win_gravity = GDK_GRAVITY_SOUTH_WEST;
  skbar.bar_height = height;
  
  /* create softkey window */
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_realize (window);
  
  gdk_window_set_type_hint (window->window, GDK_WINDOW_TYPE_HINT_TOOLBAR);
  gdk_window_set_type_hint (window->window, GDK_WINDOW_TYPE_HINT_DOCK);
  gdk_window_set_decorations (window->window, 0);
  gdk_window_set_skip_taskbar_hint (window->window, TRUE);
  gdk_window_set_skip_pager_hint (window->window, TRUE);
  gdk_window_set_geometry_hints (window->window, &geometry, GDK_HINT_WIN_GRAVITY);
  gtk_window_move (GTK_WINDOW(window), 0, gdk_screen_height() - height);
  gtk_window_set_default_size (GTK_WINDOW (window), width, height);
  gtk_window_set_title (GTK_WINDOW (window), "Softkey Area");
  gtk_window_set_keep_above (GTK_WINDOW (window), TRUE);
  gtk_window_set_accept_focus (GTK_WINDOW (window), FALSE); /* essential for softkey handling */  
  softkey_area = gtk_fixed_new ();
  for (i = 0; i < skbar.num_softkeys; i++)
      add_softkey (&skbar.softkeys[i], GTK_FIXED (softkey_area), i);
  
  gtk_container_add (GTK_CONTAINER (window), softkey_area);  
    
  setting_get_string (setting, KEY_CURRENT_THEME, &current_theme);
  if (current_theme == NULL)
    current_theme = g_strdup (DEFAULT_THEME);
   
  key = g_strconcat (THEME_PREFIX, current_theme, KEY_THEME_BG, NULL);
  g_free (current_theme);
  setting_get_string (setting, key, &bg_file);
  g_free (key); 
  
  if (bg_file)
    {
      render_bg (bg_file, window);
      g_free (bg_file);
    }
  else
      render_bg (DEFAULT_THEME_BG, window);
  
  return window;    
}
